PROGRAM monkey;

{      MONKEY  Version 1.0

    This program uses cumulative selection to produce a target string
(e.g. the alphabet in alphabetical order) through random changes.  For more
information on this subject, read the text files, MONKEY.DOC and MPROBS.DOC,
the source code of MONKEY, and Chapter 3 of Richard Dawkins' book, _The Blind
Watchmaker_.



}


{$E+,N+}

USES  DOS, CRT;

CONST
  year = 365.2425 * 24.0 * 3600.0;   { year in seconds }
  test_msg = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';  { Alphabet test message }
  max_num_copies = 700;                 { Upper Limit of number of "children" }
  first_child = 1;                      { index of first "child" }
  alphabet_pool_size = 26;          { number of letters used in alphabet test }
  user_pool_size = 27;               { number of letters in user-entered test }
  valid_set : SET OF CHAR = [' ','A'..'Z','a'..'z'];  { valid characters for
                                        user-entered test (invalid characters
                                        get converted to ' ') }

{ Pool of letters picked from randomly (' ' not used in alphabet test) }
  letter_pool : ARRAY[0..user_pool_size-1] OF CHAR = ('A','B','C','D',
                      'E','F','G','H','I','J','K','L','M','N','O','P',
                      'Q','R','S','T','U','V','W','X','Y','Z',' ');

TYPE
  real_type = Extended;               { 8087 type:  3.4E-4932 .. 1.1E4932 }
  target_type = (alphabet,entered);   { Indicates whether target string is the
                                        alphabet or a user-entered string }
  string80 = STRING[80];              { the string type }
  time_rec = RECORD                   { Time Record Type }
    hour,
    min,
    sec,
    sec100  :INTEGER;
  END;


VAR
  start_time,                         { Start and Stop times and duration  }
  stop_time,                          {   for measuring elapsed time       }
  duration   :time_rec;
  tries,                           { Reciprocal of single-step probability }
  years      :real_type;        { time in years to make all possible tries }
  target     :target_type;      { indicates type of target string }
  option,                       { single-step or cumulative selection (all
                                    change or all save one) }
  method     :BYTE;             { change single letter or test each letter
                                    position to change }
  pool_size  :BYTE;             { current size of letter pool }
  num_copies :INTEGER;          { selected number of offspring per generation }
  mut_prob   :REAL;             { probability that a given letter position
                                    will change }
  num_changes :BYTE;            { number of changes in each offspring }
  quit_flag  :BOOLEAN;          { indicates selection of QUIT PROGRAM option }

  msg,                          { Target String }
  user_msg   :string80;         { User entered string }
  msg_size   :BYTE;             { Length of Target String }

  s          :ARRAY[0..max_num_copies] OF string80;  { Mutated copies }
  diffs      :ARRAY[0..max_num_copies] OF INTEGER;   { Differences between the
                                                       copies and the target }




(* MATH SUPPORT *)

{ Constants and Variables for LOG and POWER }

CONST
  max_val = 1.0E30;         { Within range of REAL type }
  min_val = 1.0E-30;
  big_e = max_val;
  little_e = min_val;

VAR
  big_ln,                   { LN(max_val)  }
  little_ln  :REAL;         { LN(min_val)  }


FUNCTION log(v:real_type):real_type;

{  LOG returns the common log (base 10) of the value, V.  It works for the
   defined REAL_TYPE (Real, Double, or Extended).  The extra testing and
   processing is to get around the range limitations of the LN function. }

VAR
  i,
  count :INTEGER;
  e,
  ln_e,
  x     :real_type;

BEGIN
  IF v = 0.0
    THEN
      x := -max_val
    ELSE
      BEGIN
      IF v > max_val
        THEN
          BEGIN
          e := big_e;
          END
        ELSE
          IF v < min_val
            THEN
              BEGIN
              e := little_e;
              END
            ELSE
              e := 1.0;
      ln_e := LN(e);
      count := 0;
      WHILE (v < min_val) OR (v > max_val) DO
        BEGIN
        INC(count);
        v := v / e;
        END;
      x := LN(v);
      FOR i:=1 TO count DO
        x := x + ln_e;
      END;
  log := x;
END;


FUNCTION power(v:real_type;expon:INTEGER):real_type;

{  POWER takes the value, V, and raises it to the EXPON power, the value of
   which it returns.  It works for the defined REAL_TYPE (Real, Double, or
   Extended).  The extra testing and processing is to get around the range
   limitations of the LN and the EXP function. }

VAR
  i,
  count :INTEGER;
  e,
  ln_e,
  x     :real_type;

BEGIN
  IF v = 0.0
    THEN
      x := 0.0
    ELSE
      BEGIN
      x := expon * log(v);
      IF x > big_ln
        THEN
          BEGIN
          e := big_e;
          END
        ELSE
          IF x < little_ln
            THEN
              BEGIN
              e := little_e;
              END
            ELSE
              e := 1.0;
      ln_e := LN(e);
      count := 0;
      WHILE (x < little_ln) OR (x > big_ln) DO
        BEGIN
        INC(count);
        x := x - ln_e;
        END;
      x := EXP(x);
      FOR i:=1 TO count DO
        x := x * e;
      END;
  power := x;
END;


(* MAN-MACHINE INTERFACE (MMI) *)


PROCEDURE get_options;

{  GET_OPTIONS presents the Main Menu in which the user can select the
   parameters for the test.  Single key-stroke entry of the item number
   steps the selected item to its next value.  Once all parameters are
   set, pressing "X" starts the test.  Pressing "Q" terminates the program. }

VAR
  i, iy   :BYTE;
  ch,
  sel     :CHAR;
  sel_set :SET OF CHAR;
  flag,
  go_flag :BOOLEAN;

BEGIN
  go_flag := FALSE;
  quit_flag := FALSE;
  REPEAT
    ClrScr;
    WRITELN('MONKEY');
    WRITELN;
    WRITE('1. Target:  ');
    IF target = alphabet
      THEN  WRITELN('Alphabet')
      ELSE
        BEGIN
        WRITELN('User Entered -- ');
        WRITELN('"',user_msg,'"');
        END;
    WRITELN;
    WRITE('2. Option:  ');
    CASE option OF
      1 : WRITELN('Single-Step Selection');
      2 : WRITELN('Cumulative Selection -- All Offspring Change');
      3 : WRITELN('Cumulative Selection -- One Offspring Does Not Change');
    END;  {CASE}
    WRITELN;
    sel_set := ['1','2','X','Q'];
    IF option <> 1
      THEN
        BEGIN
        WRITE('3. Method:  ');
        CASE method OF
          1 : WRITELN('Change Single Letter');
          2 : WRITELN('Change Multiple Letters');
          3 : WRITELN('Chance of Changing Any or All Letters');
        END;  {CASE}
        WRITELN;
        WRITELN('4. Number of Offspring:  ',num_copies);
        WRITELN;
        sel_set := sel_set + ['3','4'];
        CASE method OF
          2 :
            BEGIN
            WRITELN('5. Number of Changes per Offspring:  ',num_changes:8);
            WRITELN;
            sel_set := sel_set + ['5'];
            END;
          3 :
            BEGIN
            WRITELN('5. Base Probability of Change:  ',mut_prob:8:4);
            WRITELN;
            sel_set := sel_set + ['5'];
            END;
          ELSE
            { do nothing }
        END;  {CASE}
        END;
    WRITELN;
    WRITELN('X.  Exit (and Run Program)');
    WRITELN;
    WRITELN('Q.  Quit');
    WRITELN;
    WRITELN;
    WRITE('Enter Choice:  ');
    REPEAT
      sel := UpCase(ReadKey);
    UNTIL sel IN sel_set;
    GoToXY(1,WhereY);
    ClrEOL;
    CASE sel OF
      '1' : IF target = alphabet
              THEN
                BEGIN
                target := entered;
                iy := WhereY;
                GoToXY(1,iy);
                IF Length(user_msg) = 0
                  THEN
                    BEGIN
                    WRITELN('No Message has been Entered');
                    WRITELN;
                    flag := TRUE;
                    END
                  ELSE
                    BEGIN
                    WRITELN('Message is:');
                    WRITELN('"',user_msg,'"');
                    WRITE('Do You Wish to Enter a Different Message? (Y/N) ');
                    REPEAT
                      ch := UpCase(ReadKey);
                    UNTIL ch IN ['Y','N'];
                    flag := (ch = 'Y');
                    END;
                IF flag
                  THEN
                    BEGIN
                    GoToXY(1,WhereY);
                    ClrEOL;
                    WRITELN('Enter New Message:');
                    READLN(user_msg);
                    FOR i:=1 TO Length(user_msg) DO     { Convert user string }
                      IF user_msg[i] IN valid_set       { to CAPs and blank   }
                        THEN  user_msg[i] := UpCase(user_msg[i])   { out      }
                        ELSE  user_msg[i] := ' ';       { invalid characters  }
                    END;
                END
              ELSE
                target := alphabet;
      '2' : BEGIN
              INC(option);
              IF option > 3
                THEN  option := 1;
            END;
      '3' : BEGIN
              INC(method);
              IF method > 3
                THEN  method := 1;
            END;
      '4' : BEGIN
              GoToXY(10,20);
              WRITE('Enter Number of Offspring:  ');
              READLN(num_copies);
            END;
      '5' : CASE method OF
              2:  BEGIN
                    GoToXY(10,WhereY);
                    WRITE('Enter Number of Changes per Offspring:  ');
                    READLN(num_changes);
                  END;
              3:  BEGIN
                    GoToXY(10,WhereY);
                    WRITE('Enter Base Probability for Change:  ');
                    READLN(mut_prob);
                  END;
              ELSE
                {do nothing}
            END;  {CASE}
      'Q' : quit_flag := TRUE;
      'X' : go_flag := TRUE;
    END;  {CASE}
  UNTIL go_flag OR quit_flag;
END;


PROCEDURE get_elapsed_time(start:time_rec;VAR duration:time_rec);

{  GET_ELAPSED_TIME reads the stop time and calculates the elapsed time in
   hours, minutes, seconds, and 1/100's of seconds.  }

  PROCEDURE adjust_time(VAR hi_unit,lo_unit:INTEGER;factor:INTEGER);

  BEGIN
    IF lo_unit < 0
      THEN
        BEGIN
        INC(lo_unit,factor);
        DEC(hi_unit);
        END;
  END;

BEGIN
  WITH stop_time DO
    GetTime(WORD(hour),WORD(min),WORD(sec),WORD(sec100));
  WITH duration DO
    BEGIN
    hour := stop_time.hour - start.hour;
    min := stop_time.min - start.min;
    sec := stop_time.sec - start.sec;
    sec100 := stop_time.sec100 - start.sec100;
    adjust_time(sec,sec100,100);
    adjust_time(min,sec,60);
    adjust_time(hour,min,60);
    END;
END;


PROCEDURE report(loops:LongInt;duration:time_rec);

{  REPORT displays the statistics of the run:  number of iterations, the
   amount of time test had run, and the equivalent number of tries and time
   required in years if single-step selection had been used.             }

BEGIN
  WRITELN;
  WITH duration DO
    BEGIN
    WRITE(loops,' iterations in ',hour:2,':');
    IF min < 10
      THEN  WRITE('0',min:1,':')
      ELSE  WRITE(min:2,':');
    IF sec < 10
      THEN  WRITE('0',sec:1,'.')
      ELSE  WRITE(sec:2,'.');
    IF sec100 < 10
      THEN  WRITE('0',sec100:1)
      ELSE  WRITE(sec100:2);
    WRITELN;
    tries := power(pool_size,msg_size);
    years := (hour * 3600.0 + min * 60.0 + sec + sec100 / 100) / year;
    years := years * tries / loops;
    WRITELN;
    WRITELN('Equivalent time for Single Step Method:');
    WRITELN(tries,' attempts in ',years,' years');
    END;
END;


(* PROGRAM SUPPORT *)


PROCEDURE get_target;

{  GET_TARGET sets the target string to either the alphabet or to the
   user-entered string.  If the alphabet test is being run, then only the
   characters "A" to "Z" are used.  If the user-entered string is used, then
   blanks are also used. }

VAR
  i  :BYTE;

BEGIN
  IF target = alphabet
    THEN
      BEGIN
      msg := test_msg;
      pool_size := alphabet_pool_size;
      END
    ELSE
      BEGIN
      msg := user_msg;
      pool_size := user_pool_size;
      END;
  msg_size := LENGTH(msg);
END;



PROCEDURE setup_screen;

{  SETUP_SCREEN clears the screen and displays the target string with its
   length at the top and instructions to exit at the bottom.  }

BEGIN
  ClrScr;
  WRITELN('"',msg,'"');
  WRITELN(msg_size);
  GoToXY(1,25);
  WRITE('Press ALT key to Exit');
  WINDOW(1,4,80,23);
  ClrScr;
END;


FUNCTION diff(s:string80):INTEGER;

{  DIFF measures the difference between the string S and the target by two
   means.  First, a count is kept of how many letters in S are different from
   the corresponding letter in the target.  If all letters are different, then
   a summation of the letter-by-letter difference between S and the target is
   calculated.

   NOTE:  Only in the single-step test have I ever seen the summation
            returned.  }

VAR
  sum,
  count  :INTEGER;
  i      :BYTE;

BEGIN
  sum := 0;
  count := 0;
  FOR i:=1 TO msg_size DO
    IF msg[i] <> s[i]
      THEN
        BEGIN
        INC(count);
        INC(sum,ABS(ORD(msg[i]) - ORD(s[i])));
        END;
  IF count = msg_size
    THEN  diff := sum
    ELSE  diff := count;
END;


FUNCTION min_diff:BYTE;

{  MIN_DIFF returns the index of the copy that differs least from the target
   string.  If two or more strings tie, then the first one found is selected. }

VAR
  i, index  :INTEGER;

BEGIN
  index := first_child;
  FOR i:=first_child+1 TO num_copies DO
    IF diffs[index] > diffs[i]
      THEN  index := i;
  min_diff := index;
END;


PROCEDURE initialize_string;

{  Creates an initial random string.  There is no guarantee that ALL letters
   are different from the corresponding letter in the target.  }

VAR
  i  :BYTE;

BEGIN
  s[0] := '';
  FOR i:=1 TO msg_size DO
    s[0] := s[0] + letter_pool[Random(pool_size)];
  diffs[0] := diff(s[0]);
END;


(* THE MAIN ROUTINES *)


PROCEDURE spawn;

{  SPAWN copies the "parent" string (index 0) into each "offspring" string and
   then performs the selected "mutation".

    NOTE:  in each case, the letter position to be changed is selected at
           random AND the letter to be placed there is also selected at
           random.   }

VAR
  i, j, k :INTEGER;

BEGIN
  FOR i:=1 TO num_copies DO
    BEGIN
    s[i] := s[0];                 { Copy the "parent" }
    IF (option <> 3) OR (i > 1)   { Do not change if first child and option }
      THEN                        {    is 3 (one child remains unchanged) }
        CASE method OF
          1 : s[i,(Random(msg_size)+1)] := letter_pool[Random(pool_size)];
          2 : FOR k:=1 TO num_changes DO
                  s[i,(Random(msg_size)+1)] := letter_pool[Random(pool_size)];
          3 : FOR j:=1 TO msg_size DO
                IF Random < mut_prob
                  THEN
                    s[i,j] := letter_pool[Random(pool_size)];
        END;  {CASE}
    diffs[i] := diff(s[i]);
    END;  {FOR i:=1 TO num_copies DO}
  i := min_diff;                  { find offspring closest to target }
  s[0] := s[i];                   { make the offspring the next parent }
  diffs[0] := diffs[i];           { save its difference for display  }
END;


PROCEDURE single_selection;

{  SINGLE_SELECTION tries to generate the target string through single-step
   selection (i.e. repeatedly try to generate the target string from scratch).
   After setting up the screen and intializing the variables, control enters
   into a REPEAT loop which increments the loop counter, generates the random
   string, and compares it to the target string.  The smallest difference,
   i.e. the closest approach to the target, is retained in CLOSEST for display
   upon exiting the loop.  Control exits the loop either when the target
   string has been generated or the user presses the ALT key.  The results are
   reported and control returns to the Main Menu.   }

VAR
  closest :INTEGER;
  loops   :LongInt;
  ch      :CHAR;
  i       :BYTE;

BEGIN
  get_target;
  setup_screen;
  closest := MAXINT;
  loops := 0;
  WITH start_time DO
    GetTime(WORD(hour),WORD(min),WORD(sec),WORD(sec100));
  REPEAT
    INC(loops);
    WRITE(loops:6,'.  ');
    initialize_string;
    WRITELN(s[0],diffs[0]:6);
    IF closest > diffs[0]
      THEN  closest := diffs[0];
  UNTIL (diffs[0] = 0) OR ((MEM[0000:$0417] AND 8) <> 0);
  get_elapsed_time(start_time,duration);
  WRITELN;
  WRITELN('Closest Approach = ',closest);
  report(loops,duration);
  WINDOW(1,1,80,25);
  GoToXY(1,25);
  ClrEOL;
  WRITE('Press any key to continue ');
  ch := ReadKey;
  WRITELN;
END;


PROCEDURE cumulative_selection;

{  CUMULATIVE_SELECTION tries to generate the target string through cumulative
   selection (i.e. using a parent string, generate multiple copies which differ
   slightly from the parent, and use the copy that comes closest to the target
   as the next parent).
   After setting up the screen and intializing the variables, control enters
   into a REPEAT loop which increments the loop counter and spawns the next
   generation.  The new parent string and its difference from the target
   string are displayed each time.  Control exits the loop either when the
   target string has been generated or the user presses the ALT key at which
   time the results are reported.  The user then chooses either to run the
   same test again or to return to the Main Menu.   }

VAR
  loops  :LongInt;
  ch     :CHAR;
  iy     :BYTE;

BEGIN
  REPEAT
    get_target;
    setup_screen;
    initialize_string;
    iy := WhereY;
    loops := 0;
    WITH start_time DO
      GetTime(WORD(hour),WORD(min),WORD(sec),WORD(sec100));
    REPEAT
      INC(loops);
      spawn;
      GoToXY(1,iy);
      WRITELN('Iterations:  ',loops);
      WRITELN;
      WRITELN(s[0]);
      WRITELN;
      WRITELN('Differs by: ',diffs[0]:4);
    UNTIL (diffs[0] = 0) OR ((MEM[0000:$0417] AND 8) <> 0);
    get_elapsed_time(start_time,duration);
    WRITELN;
    report(loops,duration);
    WINDOW(1,1,80,25);
    GoToXY(1,25);
    ClrEOL;
    WRITE('Press Space Bar to Exit; Press any other key to Repeat');
    ch := ReadKey;
    WRITELN;
  UNTIL ch = ' ';
END;


{MAIN}

BEGIN

{  Initialize values for LOG and POWER  }

  big_ln := LN(big_e);
  little_ln := LN(little_e);

{  Initialize options to default to the Alphabet Test  }

  option := 2;
  target := alphabet;
  user_msg := '';
  method := 1;
  num_copies := 100;
  mut_prob := 0.1;
  num_changes := 2;

  { Main Loop }

  Randomize;
  REPEAT
    get_options;
    IF NOT quit_flag
      THEN
        BEGIN
        IF option = 1
          THEN
            single_selection
          ELSE
            cumulative_selection;
        END;
  UNTIL quit_flag;
END.
